type param_t = Id of string | Lon of string | Lat of string | OtherParam of (string * string)
type way_tag_t = Nd of string | OtherTag of (string * (string * string) list)
type memparam_t =  Ref of string | Type of string	| OtherMem of (string * string)
type rel_tag_t = 	Member of memparam_t list	| OtherRelTag of (string * (string * string) list)

let oc = ref stdout
let create_outfile fname = oc := open_out_bin fname

let outlen = 4*1024*1024
let outbuf = Buffer.create outlen
let write s = 
	Buffer.add_string outbuf s;
	if Buffer.length outbuf > outlen then begin
		output_string !oc (Buffer.contents outbuf);
		Buffer.clear outbuf
	end

let close_outfile () = 
	output_string !oc (Buffer.contents outbuf);
	close_out !oc
	
let open_osm_old fname = 
	let ic = open_in fname in
	let lexbuf = Lexing.from_channel ic in
	lexbuf, ic
	
let inlen = 4*1024*1024
let inbuf = String.create inlen
let inpos = ref 0
let insize = ref 0 	
let ic = ref stdin

let reader s n =
	if !inpos = !insize then begin
		insize := input !ic inbuf 0 inlen;
		let nchars = min n !insize in
		String.blit inbuf 0 s 0 nchars;
		inpos := nchars;
		nchars  
	end else begin
		let nchars = min n (!insize - !inpos) in
		String.blit inbuf !inpos s 0 nchars;
		inpos := !inpos + nchars;
		nchars
	end
	
let open_osm fname = 
	ic := open_in fname;
	let lexbuf = Lexing.from_function reader in
	lexbuf, !ic

let save_params buf params =
	List.iter (fun (p, v) -> Buffer.add_char buf ' '; Buffer.add_string buf p; Buffer.add_char buf '='; Buffer.add_string buf v) params

let save_params2 buf params =
	List.iter (function OtherParam(name, value) -> 
			Buffer.add_char buf ' ';
			Buffer.add_string buf name; Buffer.add_char buf '='; Buffer.add_string buf value | _ -> ()) params

let save_tag buf (name, params) =
	Buffer.add_string buf "  <"; Buffer.add_string buf name; 
	save_params buf params;
	Buffer.add_string buf "/>\n"

let save_mem_params buf ps =
	List.iter (function Ref s -> Buffer.add_string buf " ref="; Buffer.add_string buf s
	  | Type s -> Buffer.add_string buf " type="; Buffer.add_string buf s
		| OtherMem(n,v) -> Buffer.add_char buf ' '; Buffer.add_string buf n; Buffer.add_char buf '='; Buffer.add_string buf v) ps  

let write_xml_head params = 
	let buf = Buffer.create 100 in
	Buffer.add_string buf "<?xml";
	save_params buf params;
	Buffer.add_string buf "?>\n";
	write (Buffer.contents buf)
	
let write_maintag t =
	let buf = Buffer.create 512 in
	save_tag buf t; 
	write (Buffer.contents buf)
	
let write_osmbegin params =
	let buf = Buffer.create 100 in
	Buffer.add_string buf "<osm";
	save_params buf params;
	Buffer.add_string buf ">\n";
	write (Buffer.contents buf)

let write_relation params mems smyid =
	let buf = Buffer.create 512 in
	Buffer.add_string buf "  <relation id="; Buffer.add_string buf smyid;
	save_params2 buf params;
	Buffer.add_string buf ">\n";
	List.iter (function 
		| Member ps -> Buffer.add_string buf "   <member";
			             save_mem_params buf ps;  
									 Buffer.add_string buf "/>\n"
			| OtherRelTag t -> save_tag buf t) mems;
	Buffer.add_string buf "  </relation>\n";
	write (Buffer.contents buf)	

let run f list =
	let lst = List.rev !list in
	list := [];
	List.iter f lst
		
module type Callbacks = sig	
	val xml_head : (string * string) list -> unit
  val osmbegin : (string * string) list -> unit
	val end_osm : unit -> unit
  val maintag : string * (string * string) list -> unit
	val node : param_t list -> (string * (string * string) list) list -> unit
  val way : param_t list * way_tag_t list -> unit
  val relation : param_t list * rel_tag_t list -> unit	
end