open ExtLib
open Commons
open Geo_base

let kept_points = Mem.create 987987
let kept_ways = Mem.create 987987
let kept_rels = Mem.create 9987

module FirstPass = struct
	let xml_head ps = ()
	let osmbegin ps = ()
	let maintag (name, ps) = ()
	let node nodeparams nodetags =
		let sx = ref "" and sy = ref "" and sid = ref "" in
		List.iter (function Id s -> sid := s | Lon s -> sx := s | Lat s -> sy := s | OtherParam _ -> ()) nodeparams;
		let x = float_of_string (strip !sx) and y = float_of_string (strip !sy) and id = int_of_string (strip !sid) in
		let inside = if Poly.is_inside x y then 1 else 0 in
		Mem.set kept_points id inside
	
	let postponed_ways = ref []
	let postponed_rels = ref []
			
	let way (params, tags) =
		try 
			let sid = ref "" in
			List.iter (function Id s -> sid := s | _ -> ()) params;
			let nodes = List.filter_map (function Nd sref -> Some(int_of_string (strip sref)) | OtherTag _ -> None) tags in
			let has_good_points = List.fold_left (fun has id -> has lor ((Mem.get kept_points id) land 1)) 0 nodes in
			if has_good_points <> 0 then List.iter (fun id -> 
				let old = Mem.get kept_points id in Mem.set kept_points id (old lor 2)) nodes;
			Mem.set kept_ways (int_of_string (strip !sid)) has_good_points
		with Postpone -> postponed_ways := (params, tags) :: ! postponed_ways
	
	let relation (params, tags) = 
		try 
			let smyid = ref "" in
			List.iter (function Id s -> smyid := s | _ -> ()) params;
			let myid = int_of_string (strip !smyid) in
			let keep = List.exists (function OtherRelTag _ -> false
		  	| Member ps -> 
						let sid = ref "" and mtype = ref "" in
						List.iter (function Ref s -> sid := s | Type s -> mtype := s | OtherMem _ -> ()) ps;					
						let id = int_of_string (strip !sid) in
						match !mtype with
							| "\"node\"" -> Mem.get kept_points id > 0
							| "\"way\"" -> Mem.get kept_ways id > 0
							| "\"relation\"" -> if id = myid then false else Mem.get kept_rels id > 0
							| _ -> failwith ("bad member type: " ^ !mtype)) tags in						
			Mem.set kept_rels myid (if keep then 1 else 0)
		with Postpone -> postponed_rels := (params, tags) :: !postponed_rels	

	let end_osm () = 
		run way postponed_ways;
		run relation postponed_rels							
end

module SecondPass = struct
	let xml_head params = write_xml_head params
  let maintag t = write_maintag t
  let osmbegin params = write_osmbegin params
	let node nodeparams nodetags =
	 try
		let sx = ref "" and sy = ref "" and sid = ref "" in
		List.iter (function Id s -> sid := s | Lon s -> sx := s | Lat s -> sy := s | OtherParam _ -> ()) nodeparams;
		let id = int_of_string (strip !sid) in
		if Mem.get kept_points id > 0 then begin
			let buf = Buffer.create 512 in
			Buffer.add_string buf "  <node id="; Buffer.add_string buf !sid;
			Buffer.add_string buf " lat=";       Buffer.add_string buf !sy;
			Buffer.add_string buf " lon=";       Buffer.add_string buf !sx;
			save_params2 buf nodeparams;
			if nodetags = [] then Buffer.add_string buf "/>\n" else begin
				Buffer.add_string buf ">\n";
				List.iter (save_tag buf) nodetags;
				Buffer.add_string buf "  </node>\n"	
			end;
			write (Buffer.contents buf)
		end 
   with Postpone -> ()
	
	let way (params, tags) =
	 try	
		let sid = ref "" in
		List.iter (function Id s -> sid := s | _ -> ()) params;
		if Mem.get kept_ways (int_of_string (strip !sid)) <> 0 then begin
			let buf = Buffer.create 1024 in
			Buffer.add_string buf "  <way id="; Buffer.add_string buf !sid;
			save_params2 buf params;
			Buffer.add_string buf ">\n";
			List.iter (function Nd sref -> Buffer.add_string buf "   <nd ref="; Buffer.add_string buf sref; Buffer.add_string buf "/>\n"
		 		| OtherTag t -> save_tag buf t) tags;
			Buffer.add_string buf "  </way>\n";
			write (Buffer.contents buf)	
		end
	 with Postpone -> ()
			
	let relation (params, tags) =
	 try	 		
		let smyid = ref "" in
		List.iter (function Id s -> smyid := s | _ -> ()) params;
		let myid = int_of_string (strip !smyid) in
		if Mem.get kept_rels myid <> 0 then 
			let mems = List.filter (function OtherRelTag _ -> true
		  	| Member ps -> 
						let sid = ref "" and mtype = ref "" in
						List.iter (function Ref s -> sid := s | Type s -> mtype := s | OtherMem _ -> ()) ps;					
						let id = int_of_string (strip !sid) in
						match !mtype with
							| "\"node\"" -> Mem.get kept_points id <> 0
							| "\"way\"" -> Mem.get kept_ways id <> 0
							| "\"relation\"" -> if id = myid then false else Mem.get kept_rels id <> 0
							| _ -> failwith ("bad member type: " ^ !mtype)) tags in			
			write_relation params mems !smyid
		 with Postpone -> ()
		
	let end_osm () = write "</osm>"	
end