open ExtLib
open Commons
open Geo_base

let kept_points = Mem.create 987987
let kept_ways = Mem.create 987987
let kept_rels = Mem.create 9987

let is_point_inside id x y = 
	let inside = Poly.is_inside x y in
	Mem.set kept_points id (if inside then 1 else 0);
	inside

let xml_head params = write_xml_head params
let maintag t = write_maintag t
let osmbegin params = write_osmbegin params
	
let node nodeparams nodetags =
	let sx = ref "" and sy = ref "" and sid = ref "" in
	List.iter (function Id s -> sid := s | Lon s -> sx := s | Lat s -> sy := s | OtherParam _ -> ()) nodeparams;
	let x = float_of_string (strip !sx) and y = float_of_string (strip !sy) and id = int_of_string (strip !sid) in
	if is_point_inside id x y then begin
		let buf = Buffer.create 512 in
		Buffer.add_string buf "  <node id="; Buffer.add_string buf !sid;
		Buffer.add_string buf " lat=";       Buffer.add_string buf !sy;
		Buffer.add_string buf " lon=";       Buffer.add_string buf !sx;
		save_params2 buf nodeparams;
		if nodetags = [] then Buffer.add_string buf "/>\n" else begin
			Buffer.add_string buf ">\n";
			List.iter (save_tag buf) nodetags;
			Buffer.add_string buf "  </node>\n"	
		end;
		write (Buffer.contents buf)
	end 

let postponed_ways = ref []

let way (params, tags) = 
	let keep = ref 0 in
	try
		let tags' = List.filter (function OtherTag _ -> true  
			| Nd sref -> let k = Mem.get kept_points (int_of_string (strip sref)) in keep := !keep lor k; k <> 0) tags in
		let sid = ref "" in
		List.iter (function Id s -> sid := s | _ -> ()) params;
		Mem.set kept_ways (int_of_string (strip !sid)) !keep; 
		if !keep<>0 then begin
			let buf = Buffer.create 1024 in
			Buffer.add_string buf "  <way id="; Buffer.add_string buf !sid;
			save_params2 buf params;
			Buffer.add_string buf ">\n";
			List.iter (function Nd sref -> Buffer.add_string buf "   <nd ref="; Buffer.add_string buf sref; Buffer.add_string buf "/>\n"
		 	| OtherTag t -> save_tag buf t) tags';
			Buffer.add_string buf "  </way>\n";
			write (Buffer.contents buf)			 
		end
	with Postpone -> postponed_ways := (params, tags) :: ! postponed_ways  

let postponed_rels = ref []

let relation (params, tags) = 
	let keep = ref 0 in
	try 
		let smyid = ref "" in
		List.iter (function Id s -> smyid := s | _ -> ()) params;
		let myid = int_of_string (strip !smyid) in
		let mems = List.filter (function OtherRelTag _ -> true
		  | Member ps -> 
					let sid = ref "" and mtype = ref "" in
					List.iter (function Ref s -> sid := s | Type s -> mtype := s | OtherMem _ -> ()) ps;					
					let id = int_of_string (strip !sid) in
					let inside = match !mtype with
						| "\"node\"" -> Mem.get kept_points id
						| "\"way\"" -> Mem.get kept_ways id
						| "\"relation\"" -> if id = myid then 0 else Mem.get kept_rels id
						| _ -> failwith ("bad member type: " ^ !mtype) in
					keep := !keep lor inside;
					inside <> 0) tags in
		Mem.set kept_rels myid !keep;		 
		if !keep<>0 then write_relation params mems !smyid
	with Postpone -> postponed_rels := (params, tags) :: !postponed_rels

let run_postponed () =
	run way postponed_ways;
	run relation postponed_rels

let end_osm () =
	run_postponed (); 
	write "</osm>"