%{
let parse_error s =
  Printf.printf "Parse error: %s\n" s;
  flush stdout
	 
open Geo_base
module Parser(Geo : Callbacks) = struct
%}

%token Leof Llbracket Lrbracket Lslash Leq Lquestion Losm Lnode Lway Lrelation Lid Lnd Lmember Lref Ltype Llat Llon
%token<string> Lident
%token<string> Lstring 
%start osmfile 
%type <unit> osmfile
%%
osmfile       : xmlhead osmtag { Geo.end_osm () };
xmlhead       : Llbracket Lquestion Lident otherparams Lquestion Lrbracket { Geo.xml_head (List.rev $4) };

otherparams   : { [] }
              | otherparams otherparam  { $2 :: $1 };
								
otherparam    : Lident Leq Lstring { $1, $3 };							
							
osmtag        :	osmbegin maintags Llbracket Lslash Losm Lrbracket {};
osmbegin      : Llbracket Losm otherparams Lrbracket { Geo.osmbegin (List.rev $3) }; 

maintags      : {}
              | maintags maintag  {  };							

maintag       : node {  }
              | way {  }
							| relation {  }
							| othertag { Geo.maintag $1 };
								
othertag      : Llbracket Lident otherparams Lslash Lrbracket { $2, (List.rev $3) };

othertags     : { [] }
              | othertags othertag { $2 :: $1 };

node          : Llbracket Lnode nodeparams nodeend { Geo.node (List.rev $3) $4 }

nodeend       : Lrbracket othertags Llbracket Lslash Lnode Lrbracket { List.rev $2 } 
              | Lslash Lrbracket { [] }
							;

nodeparams    : nodeparam { [$1] }
              | nodeparams nodeparam { $2 :: $1 }
							;

nodeparam     : idparam { $1 }
              | Llat Leq Lstring { Lat $3 }
							| Llon Leq Lstring { Lon $3 }
							| otherparam { OtherParam $1 }
							;
							
idparam       : Lid Leq Lstring { Id $3 };

way           :	Llbracket Lway wayparams wayend { Geo.way ($3, $4) };
wayend        : Lrbracket waytags Llbracket Lslash Lway Lrbracket { List.rev $2 }
              | Lslash Lrbracket { [] }
							;				

wayparams     : wayparam { [$1] }
              | wayparams wayparam { $2 :: $1 }
							;

wayparam      : idparam { $1 }
              | otherparam { OtherParam $1 }
							;
																	
waytags       : { [] }
              | waytags waytag { $2 :: $1 }
							;
							
waytag        : ndtag { $1 }
              | othertag { OtherTag $1 }
							;

ndtag         : Llbracket Lnd Lref Leq Lstring Lslash Lrbracket { Nd $5 };							

relation      : Llbracket Lrelation wayparams relend { Geo.relation ($3, $4) };
relend        : Lrbracket reltags Llbracket Lslash Lrelation Lrbracket { List.rev $2 }
              | Lslash Lrbracket { [] }
              ;
							
reltags       : { [] }
              | reltags reltag { $2 :: $1 }
							;

reltag        : membertag { Member $1 }
              | othertag { OtherRelTag $1 }
							;
							
membertag     : Llbracket Lmember memparams Lslash Lrbracket { List.rev $3 }
              ;

memparams     : memparam { [$1] }
              | memparams memparam { $2 :: $1 }
							;
							
memparam      : Ltype Leq Lstring { Type $3 }
              | Lref Leq Lstring { Ref $3 }
							| otherparam { OtherMem $1 }
							;

%%
end (* of module *)
let osmfile lexer buf = ()