open ExtLib
open Commons

type point = { x : float; y : float }
type line = { x1 : float; y1: float; x2 : float; y2 : float }

let same_point a b = 
  abs_float(a.x -. b.x) < 1e-10 && abs_float(a.y -. b.y) < 1e-10

let read_poly_file fname = 
  let ic = open_in fname in
  let file_lines = Std.input_lines ic in
	Enum.junk file_lines; (* skip file name *)
	let mkenum f e = 
		Enum.from (fun _ -> match Enum.get e with None | Some "END" -> raise Enum.No_more_elements | Some s -> f (s, e)) in
	let to_lines point_list = (* convert list of points to list of lines *)
		let lst =	let last = List.last point_list in
							if same_point (List.first point_list) last then point_list else last::point_list in
		let rec convert res = function
			| [] | _::[] -> res
			| p1::p2::rest -> 
					let ln = if p1.x < p2.x then {x1 = p1.x; y1 = p1.y; x2 = p2.x; y2 = p2.y } 
									 else {x1 = p2.x; y1 = p2.y; x2 = p1.x; y2 = p1.y } in										
					convert (ln::res) (p2::rest) in
		convert [] lst in 
	let all_lines = file_lines 
		|> mkenum (snd >> mkenum fst >> Enum.map (fun s -> Scanf.sscanf s " %f %f" (fun x y -> { x=x; y=y }))) 
	  |> Enum.map (List.of_enum >> to_lines) |> Enum.fold (@) [] in
	close_in ic;
	all_lines |> List.filter (fun l -> abs_float(l.x1 -. l.x2) > 1e-10);;  
	
let intersects x y line =
	if x < line.x1 || x >= line.x2 then 0 else
	let ly = (x -. line.x1) /. (line.x2 -. line.x1) *. (line.y2 -. line.y1) +. line.y1 in
	if ly >= y then 1 else 0;;  

type interval = { left : float; right : float }
module Interval = struct 
	type t = interval 
	let compare a b = 
		if a.right <= b.left then -1 else
		if b.right <= a.left then 1 else 0
end  
module M = Map.Make(Interval)

let make_tree poly = 
	let xs = poly |> List.enum |> Enum.map (fun ln -> List.enum [ln.x1; ln.x2]) |> Enum.concat |> Array.of_enum in
	Array.sort compare xs;
	let len = Array.length xs - 1 in
	let rec build_map m i =
		if i >= len then m else
		let m' = if xs.(i+1) -. xs.(i) < 1e-10 then m else
						 M.add {left = xs.(i); right = xs.(i+1)} (DynArray.create ()) m in
		build_map m' (i+1) in
	let m = build_map M.empty 0 in
	List.iter (fun ln -> 
		let interval = { left = ln.x1; right = ln.x2 } in 
		M.iter (fun iv a ->
			if Interval.compare iv interval = 0 then ((*print_string "ok, ";*) DynArray.add a ln) (*else print_string "no, "*)) m) poly;
	m;; 

let is_inside_tree tree x y =
	try
		let a = M.find { left = x; right = x } tree in
		let n =	DynArray.enum a |> Enum.map (intersects x y) |> Enum.fold (+) 0 in
		n mod 2 > 0
	with Not_found -> false
		
let poly_tree = ref M.empty

let prepare poly_filename =
	poly_tree := poly_filename |> read_poly_file  |> make_tree 

let is_inside x y = is_inside_tree !poly_tree x y
